import logging
from pkg_resources import iter_entry_points


def get_entry_points(ns=None):
    entry_points = []
    for ep in iter_entry_points('yafowil.plugin'):
        if ns is not None and ep.name != ns:
            continue
        entry_points.append(ep)
    return entry_points


def get_plugin_names(ns=None):
    return list(set([_.module_name for _ in get_entry_points(ns=ns)]))


def get_resource_directory(module_name):
    result = []
    for ep in get_entry_points(ns='resourcedir'):
        if ep.module_name != module_name:
            continue
        return ep.load()()


def _get_filepaths(module_name, ns, thirdparty):    
    for ep in get_entry_points(ns=ns):
        if ep.module_name == module_name:
            return ep.load()(thirdparty=thirdparty)
    return []


def get_javascripts(module_name, thirdparty=True):
    return _get_filepaths(module_name, 'javascripts', thirdparty)


def get_stylesheets(module_name, thirdparty=True):
    return _get_filepaths(module_name, 'stylesheets', thirdparty)


class Unset(object): 
    
    def __nonzero__(self):
        return False
    
    def __str__(self):
        return ''
    
    def __len__(self):
        return 0
    
    def __repr__(self):
        return '<UNSET>'

UNSET = Unset()


def vocabulary(definition):
    """Convert different kinds of input into a list of bi-tuples, both strings.
    """
    if callable(definition):
        definition = definition() 
    if isinstance(definition, basestring):
        return [(definition, definition),]
    # dict-like
    if hasattr(definition, '__getitem__') and hasattr(definition, 'keys'):
        return [(_, definition[_]) for _ in definition.keys()]
    
    # iterable
    if hasattr(definition, '__iter__'):
        new_vocab = []
        for entry in definition:
            if isinstance(entry, basestring):
                # entry is a string
                new_vocab.append((entry, entry))
            elif hasattr(entry, '__iter__'):
                # entry is a sequence
                parts = [_ for _ in entry]
                if len(parts) > 1:
                    # take first two parts and skips others
                    new_vocab.append(entry[0:2])
                else:
                    # rare case, inner has one value only
                    new_vocab.append((entry[0], entry[0]))
        return new_vocab
    return definition


class Tag(object):

    def __init__(self, translate):
        self.translate = translate
        self.encoding = 'utf-8'
                
    def __call__(self, tag_name, *inners, **attributes):
        """Generates some xml/html tag.
            
        ``tagname``
            name of a valid tag
            
        ``inners``
            inner content of the tag. If empty a closed tag is generated
        
        ``attributes``
            attributes of the tag, leading or trailing ``_`` underscores are 
            omitted from keywords.
    
        Example::
    
            >>> tag('p', 'Lorem Ipsum.', u'Hello World!', 
            ...     class_='fancy', id='2f5b8a234ff')
            <p class="fancy" id="2f5b8a234ff">Lorem Ipsum. Hello World.</p>
        
        """
        cl = list()
        for key, value in attributes.items():
            if value is None or value is UNSET:
                continue
            value = self.translate(value)
            if not isinstance(value, unicode):
                value = str(value).decode(self.encoding)
            cl.append((key.strip('_'), value))
        attributes = u''
        if cl:
            attributes = u' %s' % u' '.join(sorted([u'%s="%s"' % _ for _ in cl]))     
        cl = list()
        for inner in inners:
            inner = self.translate(inner)
            if not isinstance(inner, unicode):
                inner = str(inner).decode(self.encoding)
            cl.append(inner)
        if not cl:
            return u'<%(name)s%(attrs)s />' % {
                'name': tag_name,
                'attrs': attributes,
            }
        return u'<%(name)s%(attrs)s>%(value)s</%(name)s>' % {
            'name': tag_name,
            'attrs': attributes,
            'value': u''.join(i for i in cl),
        }


## Deprecation message
def _deprecated_null_localization(msg):
    logging.warn("Deprecated usage of 'yafowil.utils.tag', please use the "+\
                 "tag factory on RuntimeData instead.")
    return msg

tag = Tag(_deprecated_null_localization)        


class managedprops(object):
    
    def __init__(self, *args):
        self.__yafowil_managed_props__ = args
        
    def __call__(self, func):
        func.__yafowil_managed_props__ = self.__yafowil_managed_props__
        return func


def cssid(widget, prefix, postfix=None):
    path = widget.dottedpath.replace('.', '-')
    id = "%s-%s" % (prefix, path)
    if postfix is not None:
        id = '%s-%s' % (id, postfix) 
    return id


css_managed_props = ['class', 'error_class', 'error_class_default',
                     'required_class', 'required_class_default']

def cssclasses(widget, data, classattr='class', additional=[]):
    _classes = list()
    attrs = widget.attrs
    if attrs['error_class'] and data.errors:
        if isinstance(attrs['error_class'], basestring):
            _classes.append(attrs['error_class'])
        else:
            _classes.append(attrs['error_class_default'])
    if attrs['required_class'] and attrs['required']:
        if isinstance(attrs['required_class'], basestring):
            _classes.append(attrs['required_class'])
        else:
            _classes.append(attrs['required_class_default'])
    if attrs[classattr]:
        _classes += attrs[classattr].split()
    _classes += additional
    return _classes and ' '.join(sorted(_classes)) or None


from odict import odict
from yafowil.base import factory
from yafowil.utils import (
    UNSET,
    cssid,
    cssclasses,
    css_managed_props,
    managedprops,
)


@managedprops('structural')
def compound_extractor(widget, data):
    """Delegates extraction to children.
    """
    for childname in widget:
        child = widget[childname]
        if child.attrs.get('structural'):
            for structuralchildname in child:
                structuralchild = child[structuralchildname]
                if len(structuralchild) \
                  and structuralchild.attrs.get('structural'):
                    # call compound extractor if structural child has children
                    # which are as well structural compounds.
                    compound_extractor(structuralchild, data)
                else:
                    # call extract on widget directly
                    structuralchild.extract(data.request, parent=data)
        else:
            child.extract(data.request, parent=data)
    return odict([(k, v.extracted) for k, v in data.items()])


def compound_renderer(widget, data):
    """Delegates rendering to children.
    """
    value = widget.getter
    result = u''
    for childname in widget:
        child = widget[childname]
        if child.attrs.get('structural'):
            subdata = data
            if value is not UNSET and child.getter is UNSET:
                child.getter = value
        else:
            subdata = data.get(childname, None)
            if callable(value):
                value = value(widget, data)
            if value is not UNSET and childname in value:
                if child.getter is UNSET:
                    child.getter = value[childname]
                else:
                    raise ValueError(u"Both compound and compound member "
                                     u"provide a value for '%s'" % childname)
        if subdata is None:
            result += child(request=data.request)
        else:
            result += child(data=subdata)
    return result


factory.register(
    'compound', 
    extractors=[compound_extractor], 
    edit_renderers=[compound_renderer],
    display_renderers=[compound_renderer])

factory.doc['blueprint']['compound'] = """\
A blueprint to create a compound of widgets. This blueprint creates a node. A
node can contain sub-widgets.
"""

factory.defaults['structural'] = False
factory.doc['props']['structural'] = """\
If a compound is structural, it will be omitted in the dotted-path levels and 
will not have an own runtime-data. 
"""


def hybrid_extractor(widget, data):
    """This extractor can be used if a blueprint can act as compound or leaf.
    """
    if len(widget):
        return compound_extractor(widget, data)
    return data.extracted


@managedprops('id', *css_managed_props)
def div_renderer(widget, data):
    attrs = {
        'id': widget.attrs.get('id'),
        'class_': cssclasses(widget, data),
    }
    if len(widget):
        rendered = compound_renderer(widget, data)
    else:
        rendered = data.rendered
    return data.tag('div', rendered, **attrs)


factory.register(
    'div',
    extractors=[hybrid_extractor],
    edit_renderers=[div_renderer],
    display_renderers=[div_renderer])

factory.doc['blueprint']['div'] = """\
Like ``compound`` blueprint but renders within '<div>' element.
"""


@managedprops('legend', *css_managed_props)
def fieldset_renderer(widget, data):
    fs_attrs = {
        'id': cssid(widget, 'fieldset'),
        'class_': cssclasses(widget, data)
    }
    rendered = data.rendered
    if widget.attrs['legend']:
        rendered = data.tag('legend', widget.attrs['legend']) + rendered
    return data.tag('fieldset', rendered, **fs_attrs)   


factory.register(
    'fieldset', 
    extractors=factory.extractors('compound'), 
    edit_renderers=factory.edit_renderers('compound') + [fieldset_renderer],
    display_renderers=factory.display_renderers('compound') + \
        [fieldset_renderer])

factory.doc['blueprint']['fieldset'] = """\
Renders a fieldset around the prior rendered output.
"""

factory.defaults['fieldset.legend'] = False

factory.defaults['fieldset.class'] = None


@managedprops('action', 'method', 'enctype', *css_managed_props)
def form_edit_renderer(widget, data):
    form_attrs = {
        'action': widget.attrs['action'],
        'method': widget.attrs['method'],
        'enctype': widget.attrs['method'] == 'post' and \
                   widget.attrs['enctype'] or None,
        'novalidate': widget.attrs['novalidate'] and 'novalidate' or None,
        'class_': cssclasses(widget, data),
        'id': 'form-%s' % '-'.join(widget.path),
    }
    if callable(form_attrs['action']):
        form_attrs['action'] =  form_attrs['action'](widget, data)
    return data.tag('form', data.rendered, **form_attrs)


def form_display_renderer(widget, data):
    return data.tag('div', data.rendered)


factory.register(
    'form', 
    extractors=factory.extractors('compound'), 
    edit_renderers=factory.edit_renderers('compound') + [form_edit_renderer],
    display_renderers=factory.display_renderers('compound') + \
        [form_display_renderer])

factory.doc['blueprint']['form'] = """\
A html-form element as a compound of widgets. 
"""

factory.defaults['form.method'] = 'post'
factory.doc['props']['form.method'] = """\
One out of ``get`` or ``post``.
""" 

factory.doc['props']['form.action'] = """\
Target web address (URL) to send the form to. 
""" 

factory.defaults['form.enctype'] = 'multipart/form-data'
factory.doc['props']['form.enctype'] = """\
Encryption type of the form. Only relevant for method ``post``. Expect one out 
of ``application/x-www-form-urlencoded`` or ``multipart/form-data``.
"""

factory.defaults['form.novalidate'] = True
factory.doc['props']['form.novalidate'] = \
"""Flag whether HTML5 form validation should be suppressed.
"""


import yafowil.utils

def register():
    import yafowil.common
    import yafowil.compound
    import yafowil.table
    import yafowil.plans

for ep in yafowil.utils.get_entry_points('register'):
    ep.load()()


from yafowil.base import factory
from yafowil.compound import (
    compound_extractor,
    compound_renderer,
    hybrid_extractor,
)
from yafowil.utils import (
    cssclasses,
    css_managed_props,
    managedprops,
)


@managedprops('id', *css_managed_props)
def table_renderer(widget, data):
    attrs = {
        'id': widget.attrs.get('id'),
        'class_': cssclasses(widget, data),
    }
    return data.tag('table', data.rendered, **attrs)


factory.register(
    'table',
    extractors=factory.extractors('compound'),
    edit_renderers=factory.edit_renderers('compound') + [table_renderer],
    display_renderers=factory.display_renderers('compound') + [table_renderer])

factory.doc['blueprint']['table'] = """\
``<table>`` compound widget for table creation.
"""


def thead_renderer(widget, data):
    return data.tag('thead', data.rendered)


factory.register(
    'thead',
    extractors=factory.extractors('compound'),
    edit_renderers=factory.edit_renderers('compound') + [thead_renderer],
    display_renderers=factory.display_renderers('compound') + [thead_renderer])

factory.doc['blueprint']['thead'] = """\
``<thead>`` compound widget for table creation.
"""


def tbody_renderer(widget, data):
    return data.tag('tbody', data.rendered)


factory.register(
    'tbody',
    extractors=factory.extractors('compound'),
    edit_renderers=factory.edit_renderers('compound') + [tbody_renderer],
    display_renderers=factory.display_renderers('compound') + [tbody_renderer])

factory.doc['blueprint']['tbody'] = """\
``<tbody>`` compound widget for table creation.
"""


@managedprops('id', *css_managed_props)
def tr_renderer(widget, data):
    attrs = {
        'id': widget.attrs.get('id'),
        'class_': cssclasses(widget, data),
    }
    return data.tag('tr', data.rendered, **attrs)


factory.register(
    'tr',
    extractors=factory.extractors('compound'),
    edit_renderers=factory.edit_renderers('compound') + [tr_renderer],
    display_renderers=factory.display_renderers('compound') + [tr_renderer])

factory.doc['blueprint']['tr'] = """\
``<tr>`` compound widget for table creation.
"""


@managedprops('id', 'rowspan', 'colspan', 'label', *css_managed_props)
def th_renderer(widget, data):
    attrs = {
        'id': widget.attrs.get('id'),
        'class_': cssclasses(widget, data),
        'colspan': widget.attrs.get('colspan'),
        'rowspan': widget.attrs.get('rowspan'),
    }
    contents = widget.attrs.get('label')
    if not contents:
        contents = data.rendered
    return data.tag('th', contents, **attrs)


factory.register(
    'th',
    edit_renderers=[th_renderer],
    display_renderers=[th_renderer])

factory.doc['blueprint']['th'] = """\
``<th>`` compound widget for table creation.
"""


@managedprops('id', 'rowspan', 'colspan', *css_managed_props)
def td_renderer(widget, data):
    attrs = {
        'id': widget.attrs.get('id'),
        'class_': cssclasses(widget, data),
        'colspan': widget.attrs.get('colspan'),
        'rowspan': widget.attrs.get('rowspan'),
    }
    if len(widget):
        rendered = compound_renderer(widget, data)
    else:
        rendered = data.rendered
    return data.tag('td', rendered, **attrs)


factory.register(
    'td',
    extractors=[hybrid_extractor],
    edit_renderers=[td_renderer],
    display_renderers=[td_renderer])

factory.doc['blueprint']['td'] = """\
``<td>`` compound widget for table creation.
"""


from yafowil.base import factory

_PLANS = (
    ('stringfield',     ['field', 'label', 'error', 'text']),
    ('numberfield',     ['field', 'label', 'error', 'number']),
    ('emailfield',      ['field', 'label', 'error', 'email']),
    ('urlfield',        ['field', 'label', 'error', 'url']),
    ('passwordfield',   ['field', 'label', 'error', 'password']),
    ('textfield',       ['field', 'label', 'error', 'textarea']),
    ('selectionfield',  ['field', 'label', 'error', 'selection']),
    ('submitfield',     ['field', 'submit']),
)

for name, blueprints in _PLANS:
    factory.register_plan(name, blueprints)


from setuptools import setup, find_packages
import os

version = '1.3.1'
shortdesc = \
'YAFOWIL - declarative, flexible html forms, framework independent.'
longdesc = open(os.path.join(os.path.dirname(__file__), 'README.rst')).read()
longdesc += open(os.path.join(os.path.dirname(__file__), 'HISTORY.rst')).read()
longdesc += open(os.path.join(os.path.dirname(__file__), 'LICENSE.rst')).read()
tests_require = ['interlude', 'lxml']

setup(name='yafowil',
      version=version,
      description=shortdesc,
      long_description=longdesc,
      classifiers=[
            'Operating System :: OS Independent',
            'Programming Language :: Python',
            'Topic :: Software Development',
            'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
            'License :: OSI Approved :: BSD License',
      ],
      keywords='html input widgets form compound',
      author='BlueDynamics Alliance',
      author_email='dev@bluedynamics.com',
      url=u'http://packages.python.org/yafowil',
      license='Simplified BSD',
      packages=find_packages('src'),
      package_dir = {'': 'src'},
      namespace_packages=['yafowil'],
      include_package_data=True,
      zip_safe=True,
      install_requires=[
          'setuptools',
          'node',
      ],
      tests_require=tests_require,
      test_suite="yafowil.tests.test_suite",
      extras_require = dict(
          test=tests_require,
      ),
      entry_points = """
      [yafowil.plugin]
      register = yafowil.loader:register
      """,
)
